/*
 * Copyright (c) 2010, 2011 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef __GLOBAL_H__
#define __GLOBAL_H__

#include <mp4v2/mp4v2.h>
#include <stdio.h>
#include <stdlib.h>

/*
 * Where to send debugging output.
 */
#ifndef DEBUGOUTPUT
#define DEBUGOUTPUT stderr
#endif

/*
 * If DEBUGGING is defined, print ERROR messages with the file name and line
 * number of invocation.  Otherwise, leave that information out.
 * Then, set wasErr to 1, and jump to the specified target.
 */
#ifdef DEBUGGING
#define ERROR(target, wasErr, args...)                        \
    do {                                                      \
      fprintf(DEBUGOUTPUT, "Error: ");                        \
      fprintf(DEBUGOUTPUT, ## args);                          \
      fprintf(DEBUGOUTPUT, " [%s:%d]\n", __FILE__, __LINE__); \
      fflush(DEBUGOUTPUT);                                    \
      wasErr = 1;                                             \
      goto target;                                            \
    } while (0)
#else
#define ERROR(target, wasErr, args...)                        \
    do {                                                      \
      fprintf(DEBUGOUTPUT, "Error: ");                        \
      fprintf(DEBUGOUTPUT, ## args);                          \
      fprintf(DEBUGOUTPUT, "\n");                             \
      fflush(DEBUGOUTPUT);                                    \
      wasErr = 1;                                             \
      goto target;                                            \
    } while (0)
#endif

/*
 * If DEBUGGING is defined, print ERROR messages with the file name and line
 * number of invocation.  Otherwise, leave that information out.
 */
#ifdef DEBUGGING
#define ERROR_QUIET(target, wasErr)                           \
    do {                                                      \
      fprintf(DEBUGOUTPUT, "Quiet error");                    \
      fprintf(DEBUGOUTPUT, " [%s:%d]\n", __FILE__, __LINE__); \
      fflush(DEBUGOUTPUT);                                    \
      wasErr = 1;                                             \
      goto target;                                            \
    } while (0)
#else
#define ERROR_QUIET(target, wasErr)                           \
    do {                                                      \
      wasErr = 1;                                             \
      goto target;                                            \
    } while (0)
#endif

/*
 * If DEBUGGING is defined, enable the DEBUG macro, which prints the given
 * message, plus the file name and line number of invocation.
 */
#ifdef DEBUGGING
#define DEBUG(args...)                                        \
    do {                                                      \
      fprintf(DEBUGOUTPUT, "Debug: ");                        \
      fprintf(DEBUGOUTPUT, ## args);                          \
      fprintf(DEBUGOUTPUT, " [%s:%d]\n", __FILE__, __LINE__); \
      fflush(DEBUGOUTPUT);                                    \
    } while (0)
#else
#define DEBUG(args...)
#endif

/*
 * If DEBUGGING is defined, check if condition passes; if not, print that
 * the assertion failed (complete with file name and line number of invocation)
 * and exit the program.
 */
#ifdef DEBUGGING
#define ASSERT(condition, args...)                              \
    do {                                                        \
      if ((condition) == 0) {                                   \
        fprintf(DEBUGOUTPUT, "Assertion failed: ");             \
        fprintf(DEBUGOUTPUT, ## args);                          \
        fprintf(DEBUGOUTPUT, " [%s:%d]\n", __FILE__, __LINE__); \
        fflush(DEBUGOUTPUT);                                    \
        abort();                                                \
      }                                                         \
    } while (0)
#else
#define ASSERT(condition, args...)
#endif

/*
 * Prints the given number of milliseconds to stdout as a human-readable time.
 */
void printMilliseconds(MP4Duration milliseconds);

/*
 * Prints the given number of seconds to stdout as a human-readable time.
 */
void printSeconds(double seconds);

/*
 * Prints a progress message about a given task to standard out.  If the task
 * is complete, a newline is included at the end.
 */
void progress(const char *task, uint32_t onTrack, uint32_t numTracks,
	      MP4SampleId done, MP4SampleId total);

#endif
