/*
 * Copyright (c) 2010, 2011 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <mp4v2/mp4v2.h>
#include <getopt.h>
#include <stdio.h>
#include "global.h"

/*
 * Lists the chapters in the given MP4 input file.  Include additional
 * information in the listing if verbose is non-zero.
 * Return 0 on success, -1 on error (and print an error message).
 */
static int listChapters(MP4FileHandle fd, int verbose);

/*
 * Parse the command line options passed to mp4Chapters.
 * inputFile will point to the relevant argument within argv, and verbose will
 * be set to 0 or 1.
 * Return 0 on success, -1 on error (and print an error message).
 */
static int parseCommandLine(int argc, char **argv, char **inputFile,
			    int *verbose);

int main(int argc, char **argv)
{
  char *inputFile;
  int verbose;
  MP4FileHandle fd = MP4_INVALID_FILE_HANDLE;
  int wasErr = 0;

  /* Parse the command line and open the input file */
  if (parseCommandLine(argc, argv, &inputFile, &verbose))
    ERROR_QUIET(isErr, wasErr);
  fd = MP4Read(inputFile);
  if (fd == MP4_INVALID_FILE_HANDLE)
    ERROR(isErr, wasErr, "Cannot open input file %s", inputFile);

  /* List the chapters in the input file */
  if (listChapters(fd, verbose))
    ERROR_QUIET(isErr, wasErr);

 isErr:
  if (fd != MP4_INVALID_FILE_HANDLE)
    MP4Close(fd);
  return wasErr ? -1 : 0;
}

static int listChapters(MP4FileHandle fd, int verbose)
{
  MP4Chapter_t *list = NULL;
  MP4Duration start, end;
  uint32_t i, numChapters;
  int wasErr = 0;  

  /* Read all of the Quicktime chapters from the file */
  if (MP4GetChapters(fd, &list, &numChapters,
		     MP4ChapterTypeQt) != MP4ChapterTypeQt)
    ERROR(isErr, wasErr, "Could not read Quicktime chapter list");

  /* Print out each chapter */
  start = end = 0;
  for (i = 0; i < numChapters; i++) {
    start = end;
    end = end + list[i].duration;
    printf("%s", list[i].title);
    if (verbose) {
      printf(" (Ch. %ju: ", (uintmax_t)(i+1));
      printMilliseconds(start);
      printf(" - ");
      printMilliseconds(end);
      printf(")\n");
    }
    else
      printf("\n");
  }

 isErr:
  if (list != NULL)
    MP4Free(list);
  return wasErr ? -1 : 0;
}

static int parseCommandLine(int argc, char **argv, char **inputFile,
			    int *verbose)
{
  int ch, i;
  int wasErr = 0;

  struct option longopts[] = {
    { "verbose", no_argument, NULL, 'v' },
    { NULL,      0,           NULL, 0   }
  };
  static const char *optstring = "v";

  /* Defaults: non-verbose */
  *verbose = 0;

  /* Process input options */
  i = 0;
  while ((ch = getopt_long(argc, argv, optstring, longopts, NULL)) != -1) {
    switch (ch) {
    case 'v':
      *verbose = 1;
      break;
    default:
      i = -1;
      break;
    }
    if (i == -1)
      break;
  }

  /* Ensure correctness of option flags */
  if (i == -1 || argc - optind != 1)
    ERROR(isErr, wasErr, "Usage: %s [-v | --verbose] INPUT_FILE", argv[0]);

  /* Process the input file argument */
  *inputFile = argv[argc-1];

 isErr:
  return wasErr ? -1 : 0;
}
