/*
 * Copyright (c) 2010, 2011 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdio.h>
#include <string.h>
#include "global.h"
#include "mp4File.h"

/*
 * The name to put in files output by this utility.
 */
#define TAG_ENCODED_NAME "mp4Join 2.0.1"

/*
 * The name of the track property used to disable secondary audio tracks in
 * Quicktime.
 */
#define TAG_SECONDARY_DISABLE "tkhd.flags"

/*
 * Allocates a new struct mp4File and its struct mp4Tracks.  Returns the
 * allocated structure on success, NULL on failure (and print an error
 * message).
 */
static struct mp4File *allocateFile(uint32_t numAudio, int allocateChapter);

/*
 * Frees a struct mp4File and its struct mp4Tracks, if any of them are
 * non-NULL.
 */
static void freeFile(struct mp4File *file);

/*
 * Copies the given filename into the filename member of the file structure.
 * Returns 0 on success, -1 on error (and prints an error message).
 */
static int copyName(const char *filename, struct mp4File *dst);

/*
 * Clones all of the properties of the given source track into a new track in
 * the destination file, saving those properties into the dstTrack structure.
 * Returns 0 on success, -1 on error (and prints an error message).
 */
static int cloneTrack(struct mp4File *dst, struct mp4Track *dstTrack,
		      struct mp4File *src, struct mp4Track *srcTrack);

/*
 * Fills in all of the information about the given track, except the track ID
 * (which this function expects to be filled in already).  Returns 0 on
 * success, -1 on error (and prints an error message).
 */
static int readTrack(MP4FileHandle fd, struct mp4Track *track);

struct mp4File *readFile(const char *filename)
{
  MP4FileHandle fd;
  uint32_t numAudio, numVideo, onAudio, num, i;
  const char *type;
  struct mp4File *ret = NULL;
  int wasErr = 0;

  /* Open the source file */
  fd = MP4Read(filename);
  if (fd == MP4_INVALID_FILE_HANDLE)
    ERROR(isErr, wasErr, "Cannot open file %s for input", filename);

  /* Count the number of tracks of each type in the source file */
  numAudio = numVideo = 0;
  num = MP4GetNumberOfTracks(fd, NULL, 0);
  for (i = 0; i < num; i++) {
    type = MP4GetTrackType(fd, i+1);
    if (type == NULL)
      ERROR(isErr, wasErr, "Could not get track type");
    else if (MP4_IS_AUDIO_TRACK_TYPE(type))
      numAudio++;
    else if (MP4_IS_VIDEO_TRACK_TYPE(type))
      numVideo++;
  }
  if (numAudio == 0)
    ERROR(isErr, wasErr, "No audio track detected in file");
  if (numVideo == 0)
    ERROR(isErr, wasErr, "No video track detected in file");
  if (numVideo > 1)
    ERROR(isErr, wasErr, "Only files with one video track are supported");

  /* Allocate the structure */
  ret = allocateFile(numAudio, 0);
  if (ret == NULL)
    ERROR_QUIET(isErr, wasErr);
  ret->fd = fd;
  ret->canWrite = 0;
  if (copyName(filename, ret))
    ERROR_QUIET(isErr, wasErr);

  /* Read the track information */
  onAudio = 0;
  for (i = 0; i < num; i++) {
    type = MP4GetTrackType(fd, i+1);
    if (type == NULL)
      ERROR(isErr, wasErr, "Could not get track type");
    else if (MP4_IS_AUDIO_TRACK_TYPE(type)) {
      ret->audio[onAudio]->id = (MP4TrackId)(i+1);
      if (readTrack(fd, ret->audio[onAudio++]))
	ERROR_QUIET(isErr, wasErr);
    }
    else if (MP4_IS_VIDEO_TRACK_TYPE(type)) {
      ret->video->id = (MP4TrackId)(i+1);
      if (readTrack(fd, ret->video))
	ERROR_QUIET(isErr, wasErr);
    }
  }

 isErr:
  if (wasErr) {
    if (fd != MP4_INVALID_FILE_HANDLE)
      MP4Close(fd);
    if (ret != NULL) {
      freeFile(ret);
      ret = NULL;
    }
  }
  return ret;
}

struct mp4File *createFile(const char *filename, struct mp4File *template)
{
  uint32_t i;
  struct mp4File *ret = NULL;
  char lang[4];
  char *name = NULL;
  uint64_t flags;
  const MP4Tags *tags = NULL;
  int wasErr = 0;

  /* Allocate the structure and open the file */
  ret = allocateFile(template->numAudio, 1);
  if (ret == NULL)
    ERROR_QUIET(isErr, wasErr);
  ret->canWrite = 1;
  ret->fd = MP4Create(filename, 0);
  if (ret->fd == MP4_INVALID_FILE_HANDLE)
    ERROR(isErr, wasErr, "Cannot create file %s", filename);
  if (copyName(filename, ret))
    ERROR_QUIET(isErr, wasErr);

  /* Clone the audio and video tracks */
  if (cloneTrack(ret, ret->video, template, template->video))
    ERROR_QUIET(isErr, wasErr);
  for (i = 0; i < ret->numAudio; i++) {
    if (cloneTrack(ret, ret->audio[i], template, template->audio[i]))
      ERROR_QUIET(isErr, wasErr);
  }

  /* Copy audio metadata (cloning does a poor job of this step) */
  for (i = 0; i < ret->numAudio; i++) {
    if (MP4GetTrackLanguage(template->fd, template->audio[i]->id,
			    lang) == true) {
      if (MP4SetTrackLanguage(ret->fd, ret->audio[i]->id, lang) == false)
	ERROR(isErr, wasErr, "Cannot set track language");
    }
    if (MP4GetTrackName(template->fd, template->audio[i]->id, &name) == true) {
      if (MP4SetTrackName(ret->fd, ret->audio[i]->id, name) == false)
	ERROR(isErr, wasErr, "Cannot set track name");
      MP4Free(name);
      name = NULL;
    }
    if (MP4GetTrackIntegerProperty(template->fd, template->audio[i]->id,
				   TAG_SECONDARY_DISABLE, &flags) == true) {
      if (MP4SetTrackIntegerProperty(ret->fd, ret->audio[i]->id,
				     TAG_SECONDARY_DISABLE, flags) == false)
	ERROR(isErr, wasErr, "Unable to set flag %s", TAG_SECONDARY_DISABLE);
    }
  }

  /* Create the chapter track, referenced to the video track */
  ret->chapter->id = MP4AddChapterTextTrack(ret->fd, ret->video->id, 0);
  if (ret->chapter->id == MP4_INVALID_TRACK_ID)
    ERROR(isErr, wasErr, "Cannot create chapter track");
  ret->chapter->timeScale = ret->video->timeScale;
  ret->chapter->numSamples = 0;
  ret->chapter->duration = 0;

  /* Add iPod and tagging metadata to file, based on template tags */
  MP4AddIPodUUID(ret->fd, ret->video->id);
  tags = MP4TagsAlloc();
  if (tags == NULL)
    ERROR(isErr, wasErr, "Out of memory");
  MP4TagsFetch(tags, template->fd);
  MP4TagsSetEncodingTool(tags, TAG_ENCODED_NAME);
  MP4TagsStore(tags, ret->fd);

 isErr:
  if (wasErr && (ret != NULL)) {
    if (ret->fd != MP4_INVALID_FILE_HANDLE)
      MP4Close(ret->fd);
    freeFile(ret);
    ret = NULL;
  }
  if (name != NULL)
    MP4Free(name);
  if (tags != NULL)
    MP4TagsFree(tags);
  return ret;
}

int closeFile(struct mp4File *file)
{
  int wasErr = 0;

  /* Error cases */
  if (file == NULL)
    return 0;
  else if (file->fd == MP4_INVALID_FILE_HANDLE) {
    freeFile(file);
    return 0;
  }

  /* Non-write-permission closing is fast */
  if (file->canWrite == 0) {
    MP4Close(file->fd);
    freeFile(file);
    return 0;
  }

  /* Closing with write permission gives updates */
  printf("Writing file to disk...");
  fflush(stdout);
  MP4Close(file->fd);
  printf("done!\n");
  printf("Optimizing file layout...");
  fflush(stdout);
  if (MP4Optimize(file->filename, NULL) == false)
    printf("done!\n");
  else
    ERROR(isErr, wasErr, "File optimization failed");

 isErr:
  return wasErr ? -1 : 0;
}

int appendAudioSample(struct mp4File *dst, uint32_t dstTrack,
		      struct mp4File *src, uint32_t srcTrack,
		      MP4SampleId sampleId)
{
  MP4SampleId newId;
  int wasErr = 0;

  /* Check arguments */
  if (dstTrack >= dst->numAudio)
    ERROR(isErr, wasErr, "Invalid destination track");
  if (srcTrack >= src->numAudio)
    ERROR(isErr, wasErr, "Invalid source track");
  if (sampleId >= src->audio[srcTrack]->numSamples || sampleId + 1 == 0)
    ERROR(isErr, wasErr, "Invalid sample ID");
  if (dst->canWrite == 0)
    ERROR(isErr, wasErr, "No write permissions on destination");

  /* Copy the sample */
  newId = MP4CopySample(src->fd, src->audio[srcTrack]->id, sampleId + 1,
			dst->fd, dst->audio[dstTrack]->id,
			MP4_INVALID_DURATION);
  if (newId == MP4_INVALID_SAMPLE_ID)
    ERROR(isErr, wasErr, "Could not copy audio sample");

  /* Update track metadata */
  dst->audio[dstTrack]->numSamples += 1;
  if (dst->audio[dstTrack]->numSamples == 0)
    ERROR(isErr, wasErr, "Integer overflow");
  dst->audio[dstTrack]->duration =
    MP4GetTrackDuration(dst->fd, dst->audio[dstTrack]->id);
  if (dst->audio[dstTrack]->duration == MP4_INVALID_DURATION)
    ERROR(isErr, wasErr, "Could not recompute track duration");

 isErr:
  return wasErr ? -1 : 0;
}

int appendVideoSample(struct mp4File *dst, struct mp4File *src,
		      MP4SampleId sampleId)
{
  return appendVideoSampleSize(dst, src, sampleId, MP4_INVALID_DURATION);
}

int appendVideoSampleSize(struct mp4File *dst, struct mp4File *src,
			  MP4SampleId sampleId, MP4Duration newDuration)
{
  MP4SampleId newId;
  int wasErr = 0;

  /* Check arguments */
  if (sampleId >= src->video->numSamples || sampleId + 1 == 0)
    ERROR(isErr, wasErr, "Invalid sample ID");
  if (dst->canWrite == 0)
    ERROR(isErr, wasErr, "No write permissions on destination");

  /* Copy the sample */
  newId = MP4CopySample(src->fd, src->video->id, sampleId + 1,
			dst->fd, dst->video->id, newDuration);
  if (newId == MP4_INVALID_SAMPLE_ID)
    ERROR(isErr, wasErr, "Could not copy video sample");

  /* Update track metadata */
  dst->video->numSamples += 1;
  if (dst->video->numSamples == 0)
    ERROR(isErr, wasErr, "Integer overflow");
  dst->video->duration = MP4GetTrackDuration(dst->fd, dst->video->id);
  if (dst->video->duration == MP4_INVALID_DURATION)
    ERROR(isErr, wasErr, "Could not recompute track duration");

 isErr:
  return wasErr ? -1 : 0;
}

int appendChapter(struct mp4File *dst, const char *name, MP4Duration duration)
{
  MP4Duration oldDuration;
  int wasErr = 0;

  /* Check arguments */
  if (duration == MP4_INVALID_DURATION)
    ERROR(isErr, wasErr, "Invalid duration");
  if (dst->canWrite == 0)
    ERROR(isErr, wasErr, "No write permissions on destination");

  /* Append the chapter */
  MP4AddChapter(dst->fd, dst->chapter->id, duration, name);

  /* Update track metadata */
  oldDuration = dst->chapter->duration;
  dst->chapter->numSamples += 1;
  dst->chapter->duration += duration;
  if (dst->chapter->numSamples == 0 || dst->chapter->duration < oldDuration ||
      dst->chapter->duration < duration)
    ERROR(isErr, wasErr, "Integer overflow");

 isErr:
  return wasErr ? -1 : 0;
}

static struct mp4File *allocateFile(uint32_t numAudio, int allocateChapter)
{
  struct mp4File *ret = NULL;
  uint32_t i;
  int wasErr = 0;

  /* Allocate the structure */
  ret = (struct mp4File *)malloc(sizeof(struct mp4File));
  if (ret == NULL)
    ERROR(isErr, wasErr, "Out of memory");
  ret->numAudio = numAudio;
  ret->audio = NULL;
  ret->video = ret->chapter = NULL;

  /* Allocate space for the audio track data */
  ret->audio = (struct mp4Track **)calloc(numAudio, sizeof(struct mp4Track *));
  if (ret->audio == NULL)
    ERROR(isErr, wasErr, "Out of memory");
  for (i = 0; i < numAudio; i++) {
    ret->audio[i] = (struct mp4Track *)malloc(sizeof(struct mp4Track));
    if (ret->audio[i] == NULL)
      ERROR(isErr, wasErr, "Out of memory");
  }

  /* Allocate space for the video track data */
  ret->video = (struct mp4Track *)malloc(sizeof(struct mp4Track));
  if (ret->video == NULL)
    ERROR(isErr, wasErr, "Out of memory");

  /* If requested, also allocate space for a chapter marker track */
  if (allocateChapter) {
    ret->chapter = (struct mp4Track *)malloc(sizeof(struct mp4Track));
    if (ret->chapter == NULL)
      ERROR(isErr, wasErr, "Out of memory");
  }

 isErr:
  if (wasErr) {
    freeFile(ret);
    ret = NULL;
  }
  return ret;
}

static void freeFile(struct mp4File *file)
{
  int i;

  if (file != NULL) {
    if (file->audio != NULL) {
      for (i = 0; i < file->numAudio; i++) {
	if (file->audio[i] != NULL)
	  free(file->audio[i]);
      }
      free(file->audio);
    }

    if (file->video != NULL)
      free(file->video);
    if (file->chapter != NULL)
      free(file->chapter);

    free(file);
  }
}

static int copyName(const char *filename, struct mp4File *dst)
{
  size_t len = 0;
  int wasErr = 0;

  while (len <= PATH_MAX) {
    if (filename[len] == '\0')
      break;
    len++;
  }
  if (len > PATH_MAX)
    ERROR(isErr, wasErr, "Filename too long");

  memcpy(dst->filename, filename, (len+1)*sizeof(char));

 isErr:
  return wasErr ? -1 : 0;
}

static int cloneTrack(struct mp4File *dst, struct mp4Track *dstTrack,
		      struct mp4File *src, struct mp4Track *srcTrack)
{
  char lang[4];
  int wasErr = 0;

  dstTrack->id = MP4CloneTrack(src->fd, srcTrack->id, dst->fd,
			       MP4_INVALID_TRACK_ID);
  if (dstTrack->id == MP4_INVALID_TRACK_ID)
    ERROR(isErr, wasErr, "Cannot clone track");
  if (MP4GetTrackLanguage(src->fd, srcTrack->id, lang) == true) {
    if (MP4SetTrackLanguage(dst->fd, dstTrack->id, lang) == false)
      ERROR(isErr, wasErr, "Could not set language");
  }
  if (readTrack(dst->fd, dstTrack))
    ERROR_QUIET(isErr, wasErr);

 isErr:
  return wasErr ? -1 : 0;
}

static int readTrack(MP4FileHandle fd, struct mp4Track *track)
{
  int wasErr = 0;

  track->timeScale = MP4GetTrackTimeScale(fd, track->id);
  track->duration = MP4GetTrackDuration(fd, track->id);
  if (track->duration == MP4_INVALID_DURATION)
    ERROR(isErr, wasErr, "Could not get track duration");
  track->numSamples = MP4GetTrackNumberOfSamples(fd, track->id);
  if (track->numSamples == MP4_INVALID_SAMPLE_ID) {
    if (track->duration != 0)
      ERROR(isErr, wasErr, "Could not get number of samples");
    else
      track->numSamples = 0;
  }

 isErr:
  return wasErr ? -1 : 0;
}
