/*
 * Copyright (c) 2010, 2011 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef __FILE_H__
#define __FILE_H__

#include <mp4v2/mp4v2.h>
#include <limits.h>
#include <stdint.h>

/*
 * The track structure represents a single track in an MP4 file.  For chapter
 * tracks (as opposed to audio or video tracks), the number of samples and
 * duration are set to zero.
 */
struct mp4Track
{
  MP4TrackId id;          /* The track ID of the track */
  uint32_t timeScale;     /* Number of ticks per second */
  MP4SampleId numSamples; /* Number of samples in the track */
  MP4Duration duration;   /* Duration of the track in timeScale ticks */
};

/*
 * The file structure represents an MP4 file with its audio tracks and video
 * track.
 */
struct mp4File
{
  MP4FileHandle fd;          /* The file descriptor for the open file */
  char filename[PATH_MAX+1]; /* The filename */
  int canWrite;              /* If there are write permissions on the file */
  struct mp4Track **audio;   /* The audio tracks */
  uint32_t numAudio;         /* The number of audio tracks */
  struct mp4Track *video;    /* The video track */
  struct mp4Track *chapter;  /* The chapter track, or NULL if canWrite == 0 */
};

/*
 * Opens the given MP4 file for reading.  Returns an allocated struct mp4File
 * on success or NULL on error (and prints an error message).  The caller is
 * responsible for freeing the structure by calling closeFile.
 */
struct mp4File *readFile(const char *filename);

/*
 * Creates the given MP4 file with write permission, based on the audio and
 * video tracks in the given template file.  Returns an allocated struct
 * mp4File on success or NULL on error (and prints an error message).  The
 * caller is responsible for freeing the structure by calling closeFile.
 */
struct mp4File *createFile(const char *filename, struct mp4File *template);

/*
 * Closes the given MP4 file.  If the file is open with write permissions,
 * optimize the file structure as it is closing (and print progress messages).
 * If the given file is NULL, this function does nothing and returns 0.
 * Return 0 on success, -1 on error (and print an error message).  Guaranteed
 * to return 0 for files without write permission.
 */
int closeFile(struct mp4File *file);

/*
 * Appends the given audio sample from the source file to the end of the given
 * audio track in the destination file.  The track numbers are indexed from
 * zero to one fewer than the number of audio tracks in the file.  Return 0 on
 * success, -1 on error (and print an error message).
 */
int appendAudioSample(struct mp4File *dst, uint32_t dstTrack,
		      struct mp4File *src, uint32_t srcTrack,
		      MP4SampleId sampleId);

/*
 * Appends the given video sample from the source file to the end of the video
 * track in the destination file.  Return 0 on success, -1 on error (and print
 * an error message).
 */
int appendVideoSample(struct mp4File *dst, struct mp4File *src,
		      MP4SampleId sampleId);

/*
 * Appends the given video sample from the source file to the end of the video
 * track in the destination file, resizing the sample to have the given
 * duration in video-track timescale.  If MP4_INVALID_DURATION is given as the
 * new duration, this call is identical to calling appendVideoSample. Return 0
 * on success, -1 on error (and print an error message).
 */
int appendVideoSampleSize(struct mp4File *dst, struct mp4File *src,
			  MP4SampleId sampleId, MP4Duration newDuration);

/*
 * Appends the given chapter to the end of the chapter track.  The duration is
 * in the timescale of the chapter track.  Return 0 on success, -1 on error
 * (and print an error message).
 */
int appendChapter(struct mp4File *dst, const char *name, MP4Duration duration);

#endif
