/*
 * Copyright (c) 2010, 2011 Ryan Vogt <vogt@cs.ualberta.ca>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <mp4v2/mp4v2.h>
#include <stdio.h>
#include <unistd.h>
#include "global.h"
#include "mp4File.h"

/*
 * Appends the given source file onto the end of the given destination file.
 * Set the number to a value > 0, indicating where in the appending queue this
 * source file is located, and set total to the total number of files being
 * appended.  Returns 0 on success, -1 on error (and prints an error message).
 */
static int appendFile(struct mp4File *dst, struct mp4File *src, int number,
		      int total);

/*
 * Comes up with a plan for how much of the source file to append into the
 * destination file, such that (if isLast == 0) the audio/video "edge" left at
 * the end of the file is as smooth as possible.  Fills in the number of audio
 * and video samples to copy for each track, as well as the new duration for
 * the last video sample (or MP4_INVALID_DURATION to leave it untouched), and
 * prints informative messages for the user.  Returns 0 on success, -1 on error
 * (and prints an error message).
 */
static int planCopy(struct mp4File *dst, struct mp4File *src, int isLast,
		    MP4SampleId *audioToCopy, MP4SampleId *videoToCopy,
		    MP4Duration *lastVideoDuration);

/*
 * Determines the number of samples to copy from the source file audio track
 * so that the copies samples will have the target duration.  The
 * resultDuration pointer is filled in with the actual resulting duration,
 * which may not be exactly the target duration.  Returns 0 on success, -1 on
 * error.
 */
static int audioAmount(struct mp4File *src, uint32_t srcTrack,
		       MP4Duration targetDuration, MP4SampleId *numSamples,
		       MP4Duration *resultDuration);

/*
 * Determines the number of samples to copy from the source file video track,
 * as well as the duration to set for the last sample, such that the copied
 * samples will have the target duration.
 * Returns 0 on success, -1 on error.
 */
static int videoAmount(struct mp4File *src, MP4Duration targetDuration,
		       MP4SampleId *numSamples,
		       MP4Duration *lastSampleDuration);

/*
 * Appends the chapters from the source file to the destination file, in such
 * a way as to take up exactly the given duration.
 * Returns 0 on success, -1 on error (and prints an error message).
 */
static int appendChapters(struct mp4File *dst, struct mp4File *src,
			  MP4Duration duration, int partNumber);

/*
 * Returns the percentage change from the original duration to the new duration.
 */
static double percentChange(MP4Duration origDur, MP4Duration newDur);

/*
 * Fills in the array of input files and the single output file, based on the
 * command-line options.  Input files returned from this function are believed
 * to be compatibile for concatenation.  Returns 0 on success, -1 on error (and
 * prints an error message).
 */
static int parseCommandLine(int argc, char **argv,
			    struct mp4File ***inputFiles, int *numInputFiles,
			    struct mp4File **outputFile);

/*
 * Verifies (to some extent) that file two can be appended to the end of file
 * one.  Returns 0 on success, -1 on error (and prints an error message).
 */
static int verifyCompatibility(struct mp4File *one, struct mp4File *two);

int main(int argc, char **argv)
{
  struct mp4File **inputFiles = NULL;
  int numInputFiles = 0;
  struct mp4File *outputFile = NULL;
  int i;
  int wasErr = 0;

  /* Parse the command line and append the files one at a time to the output */
  if (parseCommandLine(argc, argv, &inputFiles, &numInputFiles, &outputFile))
    ERROR_QUIET(isErr, wasErr);
  for (i = 0; i < numInputFiles; i++) {
    if (appendFile(outputFile, inputFiles[i], i+1, numInputFiles))
      ERROR_QUIET(isErr, wasErr);
    printf("\n");
  }

 isErr:
  if (inputFiles != NULL) {
    for (i = 0; i < numInputFiles; i++)
      closeFile(inputFiles[i]);
    free(inputFiles);
  }
  closeFile(outputFile);
  return wasErr ? -1 : 0;
}

static int appendFile(struct mp4File *dst, struct mp4File *src, int number,
		      int total)
{
  MP4SampleId *audioToCopy = NULL;
  MP4SampleId videoToCopy = 0, i;
  MP4Duration lastVideoDuration = 0, oldVideoDuration;
  uint32_t onTrack;
  int ret;
  int wasErr = 0;

  /* Decide how much audio and video to copy from the source */
  audioToCopy = (MP4SampleId *)calloc(src->numAudio, sizeof(MP4SampleId));
  if (audioToCopy == NULL)
    ERROR(isErr, wasErr, "Out of memory");
  if (planCopy(dst, src, (number == total), audioToCopy, &videoToCopy,
	       &lastVideoDuration))
    ERROR_QUIET(isErr, wasErr);
  oldVideoDuration = dst->video->duration;

  /* Copy the audio tracks from src to dst */
  for (onTrack = 0; onTrack < src->numAudio; onTrack++) {
    for (i = 0; i < audioToCopy[onTrack]; i++) {
      progress("Copying audio track", onTrack+1, src->numAudio, i,
	       audioToCopy[onTrack]);
      if (appendAudioSample(dst, onTrack, src, onTrack, i))
	ERROR_QUIET(isErr, wasErr);
    }
    progress("Copying audio track", onTrack+1, onTrack+1,
	     audioToCopy[onTrack], audioToCopy[onTrack]);
  }

  /* Copy the video track from src to dst */
  for (i = 0; i < videoToCopy; i++) {
    progress("Copying video track", 1, 1, i, videoToCopy);
    if (i == videoToCopy - 1)
      ret = appendVideoSampleSize(dst, src, i, lastVideoDuration);
    else
      ret = appendVideoSample(dst, src, i);
    if (ret)
      ERROR_QUIET(isErr, wasErr);
  }
  progress("Copying video track", 1, 1, videoToCopy, videoToCopy);

  /* Copy the chapters from the source to the destination */
  printf("Copying chapter titles...");
  fflush(stdout);
  if (appendChapters(dst, src, (dst->video->duration - oldVideoDuration),
		     number))
    ERROR_QUIET(isErr, wasErr);
  printf("done!\n");

 isErr:
  if (audioToCopy != NULL)
    free(audioToCopy);
  return wasErr ? -1 : 0;
}

static int planCopy(struct mp4File *dst, struct mp4File *src, int isLast,
		    MP4SampleId *audioToCopy, MP4SampleId *videoToCopy,
		    MP4Duration *lastVideoDuration)
{
  double duration, shortestDuration = 0.0;
  double videoStart, audioStart;
  uint32_t onTrack, shortestAudioTrack = 0;
  MP4Duration appendedShortestDur = 0, appendedFirstDur = 0;
  MP4Duration targetDur, actualDur;
  MP4Duration sampleDur;
  int wasErr = 0;

  /* Print an introductory message */
  printf("Preparing to append %s\n", src->filename);
  for (onTrack = 0; onTrack < src->numAudio; onTrack++) {
    printf("    Audio track %ju duration: ", (uintmax_t)(onTrack+1));
    duration = ((double)(src->audio[onTrack]->duration)) /
      (src->audio[onTrack]->timeScale);
    if (onTrack == 0 || duration < shortestDuration) {
      shortestDuration = duration;
      shortestAudioTrack = onTrack;
    }
    printSeconds(duration);
    printf("\n");
  }
  printf("    Video track 1 duration: ");
  printSeconds(((double)(src->video->duration))/(src->video->timeScale));
  printf("\n");

  if (isLast) {
    /*
     * If this is the last source file being appended, be lazy and just copy
     * the whole thing (no need for a smooth edge).
     */
    for (onTrack = 0; onTrack < src->numAudio; onTrack++)
      audioToCopy[onTrack] = src->audio[onTrack]->numSamples;
    *videoToCopy = src->video->numSamples;
    *lastVideoDuration = MP4_INVALID_DURATION;
  }
  else {
    /*
     * Otherwise, determine how long the audio track will be after the shortest
     * source audio track is appended to its counterpart in the destination
     * file.
     */
    audioToCopy[shortestAudioTrack] =
      src->audio[shortestAudioTrack]->numSamples;
    appendedShortestDur = dst->audio[shortestAudioTrack]->duration +
      src->audio[shortestAudioTrack]->duration;
    if (appendedShortestDur < dst->audio[shortestAudioTrack]->duration ||
	appendedShortestDur < src->audio[shortestAudioTrack]->duration)
      ERROR(isErr, wasErr, "Integer overflow");
    if (shortestAudioTrack == 0)
      appendedFirstDur = appendedShortestDur;

    /*
     * Copy an amount of audio from the other tracks to create as smooth an
     * edge as possible.
     */
    for (onTrack = 0; onTrack < src->numAudio; onTrack++) {
      if (onTrack == shortestAudioTrack)
	continue;
      targetDur =
	MP4ConvertFromTrackDuration(src->fd,
				    src->audio[shortestAudioTrack]->id,
				    appendedShortestDur,
				    src->audio[onTrack]->timeScale);
      if (targetDur == 0)
	ERROR(isErr, wasErr, "Error converting to audio track time scale");
      if (targetDur < dst->audio[onTrack]->duration)
	ERROR(isErr, wasErr, "Duration underflow");
      if (audioAmount(src, onTrack, targetDur - dst->audio[onTrack]->duration,
		      audioToCopy + onTrack, &actualDur))
	ERROR_QUIET(isErr, wasErr);
      if (onTrack == 0)
	appendedFirstDur = dst->audio[0]->duration + actualDur;
    }

    /*
     * Do the same computation for the video track, trying to synchronize it to
     * the first audio track (assuming that is the most important audio track
     * to keep synchronized to the video).
     */
    targetDur = MP4ConvertFromTrackDuration(src->fd, src->audio[0]->id,
					    appendedFirstDur,
					    src->video->timeScale);
    if (targetDur == 0)
      ERROR(isErr, wasErr, "Error converting from audio to video time scale");
    if (targetDur < dst->video->duration)
      ERROR(isErr, wasErr, "Duration underflow");
    if (videoAmount(src, targetDur - dst->video->duration, videoToCopy,
		    lastVideoDuration))
      ERROR_QUIET(isErr, wasErr);
  }

  /* Print out the plan for each audio track */
  videoStart = ((double)(dst->video->duration))/(dst->video->timeScale);
  for (onTrack = 0; onTrack < src->numAudio; onTrack++) {
    printf("Plan for audio track %ju:\n", (uintmax_t)(onTrack+1));
    printf("    Will copy %ju / %ju samples\n",
	   (uintmax_t)(audioToCopy[onTrack]),
	   (uintmax_t)(src->audio[onTrack]->numSamples));
    audioStart = ((double)(dst->audio[onTrack]->duration))/
      (dst->audio[onTrack]->timeScale);
    if (audioStart == videoStart)
      printf("    Audio and video will be synchronized\n");
    else if (audioStart < videoStart)
      printf("    Audio will be %.3lf milliseconds behind video\n",
	     (videoStart - audioStart)*1000.0);
    else
      printf("    Audio will be %.3lf milliseconds ahead of video\n",
	     (audioStart - videoStart)*1000.0);
  }

  /* Print out the plan for the video track */
  printf("Plan for video track 1:\n");
  printf("    Will copy %ju / %ju samples\n", (uintmax_t)(*videoToCopy),
	 (uintmax_t)(src->video->numSamples));
  if (*lastVideoDuration == MP4_INVALID_DURATION)
    printf("    Last video sample duration unmodified\n");
  else {
    sampleDur = MP4GetSampleDuration(src->fd, src->video->id, *videoToCopy);
    if (sampleDur == MP4_INVALID_DURATION)
      ERROR(isErr, wasErr, "Cannot get original sample duration");
    printf("    Last video sample duration: %.3lf seconds -> %.3lf seconds\n",
	   ((double)sampleDur)/(src->video->timeScale),
	   ((double)(*lastVideoDuration))/(src->video->timeScale));
  }

 isErr:
  return wasErr ? -1 : 0;
}

static int audioAmount(struct mp4File *src, uint32_t srcTrack,
		       MP4Duration targetDuration, MP4SampleId *numSamples,
		       MP4Duration *resultDuration)
{
  MP4SampleId intersectingSample;
  MP4Timestamp startTime, endTime;
  MP4Duration includeDesync, excludeDesync;
  int wasErr = 0;

  /* Find the sample ID of the sample that covers the desired time */
  intersectingSample = MP4GetSampleIdFromTime(src->fd,
					      src->audio[srcTrack]->id,
					      targetDuration, false);
  if (intersectingSample == MP4_INVALID_SAMPLE_ID)
    ERROR(isErr, wasErr, "Could not convert time to sample ID");
  if (intersectingSample > src->audio[srcTrack]->numSamples)
    intersectingSample = src->audio[srcTrack]->numSamples;

  /* Determine the start and end times of the intersecting sample */
  startTime = MP4GetSampleTime(src->fd, src->audio[srcTrack]->id,
			       intersectingSample);
  if (startTime == MP4_INVALID_TIMESTAMP)
    ERROR(isErr, wasErr, "Could not get sample start time");
  endTime = MP4GetSampleDuration(src->fd, src->audio[srcTrack]->id,
				 intersectingSample);
  if (endTime == MP4_INVALID_DURATION)
      ERROR(isErr, wasErr, "Could not get intersecting sample duration");
  endTime += startTime;

  /* Include or exclude that sample, minimizing absolute desynchronization */
  if (targetDuration > startTime)
    excludeDesync = targetDuration - startTime;
  else
    excludeDesync = startTime - targetDuration;
  if (targetDuration > endTime)
    includeDesync = targetDuration - endTime;
  else
    includeDesync = endTime - targetDuration;
  if (excludeDesync <= includeDesync) {
    *numSamples = intersectingSample - 1;
    *resultDuration = (MP4Duration)startTime;
  }
  else {
    *numSamples = intersectingSample;
    *resultDuration = (MP4Duration)endTime;
  }

 isErr:
  return wasErr ? -1 : 0;
}

static int videoAmount(struct mp4File *src, MP4Duration targetDuration,
		       MP4SampleId *numSamples,
		       MP4Duration *lastSampleDuration)
{
  MP4SampleId intersectingSample;
  MP4Timestamp startTime;
  MP4Duration originalDuration, newDuration;
  double ratio = 0.0;
  int foundApproach = 0;
  int noSecondApproach = 0;
  int wasErr = 0;

  /* Find the sample ID of the sample that covers the desired time */
  intersectingSample = MP4GetSampleIdFromTime(src->fd, src->video->id,
					      targetDuration, false);
  if (intersectingSample == MP4_INVALID_SAMPLE_ID)
    ERROR(isErr, wasErr, "Could not convert time to sample ID");
  if (intersectingSample > src->video->numSamples) {
    intersectingSample = src->video->numSamples;
    noSecondApproach = 1;
  }

  /*
   * We have two choices: we can either resize the intersecting sample, or we
   * can resize the previous sample (and not include the intersecting sample).
   * Choose whichever has a smaller multiplicative effect on the length of the
   * modified sample.
   */

  /* First, compute the effect of resizing the intersecting sample */
  startTime = MP4GetSampleTime(src->fd, src->video->id, intersectingSample);
  if (startTime == MP4_INVALID_TIMESTAMP)
    ERROR(isErr, wasErr, "Could not get sample start time");
  if (startTime < targetDuration) {
    originalDuration = MP4GetSampleDuration(src->fd, src->video->id,
					    intersectingSample);
    if (originalDuration == MP4_INVALID_DURATION)
      ERROR(isErr, wasErr, "Could not get intersecting sample duration");
    newDuration = targetDuration - startTime;
    ratio = percentChange(originalDuration, newDuration);
    *numSamples = intersectingSample;
    *lastSampleDuration = newDuration;
    foundApproach = 1;
  }

  /* Next, compute the effect of lengething the previous sample */
  if (intersectingSample > 1 && noSecondApproach == 0) {
    startTime = MP4GetSampleTime(src->fd, src->video->id,
				 intersectingSample-1);
    if (startTime == MP4_INVALID_TIMESTAMP)
      ERROR(isErr, wasErr, "Could not get sample start time");
    if (startTime < targetDuration) {
      originalDuration = MP4GetSampleDuration(src->fd, src->video->id,
					      intersectingSample-1);
      if (originalDuration == MP4_INVALID_DURATION)
	ERROR(isErr, wasErr, "Could not get previous sample duration");
      newDuration = targetDuration - startTime;
      if (foundApproach == 0 ||
	  percentChange(originalDuration, newDuration) < ratio) {
	*numSamples = intersectingSample - 1;
	*lastSampleDuration = newDuration;
	foundApproach = 1;
      }
    }
  }

  /* Ensure that one of the two approaches worked */
  if (foundApproach == 0)
    ERROR(isErr, wasErr, "Could not determine method to modify video track");

 isErr:
  return wasErr ? -1 : 0;
}

static int appendChapters(struct mp4File *dst, struct mp4File *src,
			  MP4Duration duration, int partNumber)
{
  char defaultTitle[MP4V2_CHAPTER_TITLE_MAX+1];
  MP4Chapter_t *list = NULL;
  uint32_t numChapters = 0;
  uint32_t i;
  MP4Duration chapterDuration;
  int wasErr = 0;

  /* Get the Quicktime chapter list; on error, assume there are no chapters */
  if (MP4GetChapters(src->fd, &list, &numChapters,
		     MP4ChapterTypeQt) != MP4ChapterTypeQt)
    numChapters = 0;

  /* If no chapters are found, just insert a default chapter */
  if (numChapters == 0) {
    printf("no chapters found...");
    fflush(stdout);
    snprintf(defaultTitle, MP4V2_CHAPTER_TITLE_MAX+1, "Part %d", partNumber);
    MP4AddChapter(dst->fd, dst->chapter->id, duration, defaultTitle);
    return 0;
  }

  /*
   * Otherwise, insert chapters up to the duration; expand the last chapter to
   * meet the duration if necessary.
   */
  for (i = 0; i < numChapters; i++) {
    if (i == numChapters - 1)
	chapterDuration = duration;
    else {
      /* The list structure contains duration in milliseconds */
      chapterDuration = MP4ConvertToTrackTimestamp(dst->fd, dst->chapter->id,
						   list[i].duration,
						   MP4_MILLISECONDS_TIME_SCALE);
      if (chapterDuration == 0)
	ERROR(isErr, wasErr, "Could not convert chapter duration");
      if (chapterDuration > duration)
	chapterDuration = duration;
    }
    MP4AddChapter(dst->fd, dst->chapter->id, chapterDuration, list[i].title);
    duration -= chapterDuration;
    if (duration == 0)
      break;
  }

 isErr:
  return wasErr ? -1 : 0;
}

static double percentChange(MP4Duration origDur, MP4Duration newDur)
{
  if (origDur > newDur)
    return ((double)(origDur - newDur))/origDur;
  else
    return ((double)(newDur - origDur))/origDur;
}

static int parseCommandLine(int argc, char **argv, struct mp4File ***inputFiles,
			    int *numInputFiles, struct mp4File **outputFile)
{
  int i;
  int wasErr = 0;

  /* Determine the number of input files */
  *inputFiles = NULL;
  *outputFile = NULL;
  if (argc < 4)
    ERROR(isErr, wasErr, "Usage: %s INPUT_FILE INPUT_FILE [...] OUTPUT_FILE",
	  argv[0]);
  *numInputFiles = argc - 2;
  *inputFiles = (struct mp4File **)calloc(*numInputFiles,
					  sizeof(struct mp4File *));
  if (*inputFiles == NULL)
    ERROR(isErr, wasErr, "Out of memory");

  /* Open each of the input files */
  for (i = 0; i < *numInputFiles; i++) {
    (*inputFiles)[i] = readFile(argv[i+1]);
    if ((*inputFiles)[i] == NULL)
      ERROR_QUIET(isErr, wasErr);
  }

  /* Check compatibility of input files */
  for (i = 0; i < (*numInputFiles)-1; i++) {
    if (verifyCompatibility((*inputFiles)[i], (*inputFiles)[i+1]))
      ERROR_QUIET(isErr, wasErr);
  }

  /* Ensure that the output file doesn't already exist, and create it */
  if (access(argv[argc-1], F_OK) == 0)
    ERROR(isErr, wasErr, "File %s already exists", argv[argc-1]);
  printf("Creating new file %s\n", argv[argc-1]);
  *outputFile = createFile(argv[argc-1], (*inputFiles)[0]);
  if (*outputFile == NULL)
    ERROR_QUIET(isErr, wasErr);
  printf("\n");

 isErr:
  if (wasErr) {
    if (*inputFiles != NULL) {
      for (i = 0; i < *numInputFiles; i++)
	closeFile((*inputFiles)[i]);
      free(*inputFiles);
      *inputFiles = NULL;
    }
    closeFile(*outputFile);
    *outputFile = NULL;
  }
  return wasErr ? -1 : 0;
}

static int verifyCompatibility(struct mp4File *one, struct mp4File *two)
{
  int i;
  int wasErr = 0;

  /* Verify, as far as possible, equal audio properties */
  if (one->numAudio != two->numAudio)
    ERROR(isErr, wasErr, "Input files %s and %s have a different "
	  "number of audio tracks", one->filename, two->filename);
  for (i = 0; i < one->numAudio; i++) {
    if (one->audio[i]->timeScale != two->audio[i]->timeScale)
      ERROR(isErr, wasErr, "Input files %s and %s have different "
	    "audio time scales", one->filename, two->filename);
    if (MP4GetTrackFixedSampleDuration(one->fd, one->audio[i]->id) !=
	MP4GetTrackFixedSampleDuration(two->fd, two->audio[i]->id))
      ERROR(isErr, wasErr, "Input files %s and %s have different "
	    "audio sample durations", one->filename, two->filename);
    if (MP4GetTrackEsdsObjectTypeId(one->fd, one->audio[i]->id) !=
	MP4GetTrackEsdsObjectTypeId(one->fd, one->audio[i]->id))
      ERROR(isErr, wasErr, "Input files %s and %s have different audio types",
	    one->filename, two->filename);
  }

  /* Verify, as far as possible, equal video properties */
  if (one->video->timeScale != two->video->timeScale)
    ERROR(isErr, wasErr, "Input files %s and %s have different "
	  "video time scales", one->filename, two->filename);
  if (MP4GetTrackVideoWidth(one->fd, one->video->id) !=
      MP4GetTrackVideoWidth(two->fd, two->video->id) ||
      MP4GetTrackVideoHeight(one->fd, one->video->id) !=
      MP4GetTrackVideoHeight(two->fd, two->video->id))
    ERROR(isErr, wasErr, "Input files %s and %s have different video sizes",
	  one->filename, two->filename);

 isErr:
  return wasErr ? -1 : 0;
}
